# Copyright (c) 2016 EasyStack Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections.abc import Generator
from typing import Any


def flatten_dict_to_keypairs(
    d: dict[str, Any], separator: str = ':'
) -> Generator[tuple[str, Any], None, None]:
    """Generator that produces sequence of keypairs for nested dictionaries.

    :param d: dictionaries which may be nested
    :param separator: symbol between names
    """
    for name, value in sorted(d.items()):
        if isinstance(value, dict):
            for subname, subvalue in flatten_dict_to_keypairs(
                value, separator
            ):
                yield f'{name}{separator}{subname}', subvalue
        else:
            yield name, value
